unit dataio;
(************************************************************
 ***  This contains useful procedures for reading/writing ***
 ***  encrypted data to text files.                       ***
 ***                                                      ***
 ***  PACKREC/UNPACKREC  Write a data type, key and value ***
 ***                     to a file in encrypted form      ***
 ***                                                      ***
 ***  STRINGFROMVAR      Variant -> string (dates handled)***
 ***  VARFROMSTRING      String -> Variant (dates handled)***
 ************************************************************)

interface

uses Sysutils,useful;

Type

  // keyed data file manipulation-----------------------------------
  TKeyDataFile = Class(TObject)
  { This object will read and write variants, encrypted, to a text file
    The typical sequence for reading is:
      with TKeyDataFile.Create do begin
        TRY
          nr := OpenForRead(aFileName);
          for i := 1 to nr do begin
            if ReadNext(ky,v) then <do something> else break;
          end;
        FINALLY
          free;
        END;
      end;

    The typical sequence for writing is:
      with TKeyDataFile.Create do begin
        TRY
          OpenForWrite(aFileName,NoRecs);
          for i := 1 to NoRecs do begin
            <get key and variant>
            WriteNext(ky,v);
          end;
        FINALLY
          free;
        END;
      end;

          }

    public
      Function OpenForRead(const aFileName:String):integer;
      Procedure OpenForWrite(const aFileName:String;const NoToWrite:integer);
      Procedure Close;
      Destructor Destroy; override;
      procedure WriteNext(const ky:string;v:variant);
      Function ReadNext(var ky:string;var v:variant):boolean;
    private
      F : TextFile;
      Open : boolean;
      Reading : boolean;
      nrecs : Integer;
    end;


  // general utility functions--------------------------------
  function PackRec(const vt:integer;const k,v:string):string;
  Function StringFromVar(const v:Variant):string;
  Procedure UnPackRec(const S:string;var vt:integer;var k,v:string);
  Function VarFromString(const vtype:Integer;const s:string):variant;

implementation


function PackRec(const vt:integer;const k,v:string):string;
var
  B1,B2,B3,B4 : Char;
  C :integer;
  R,i : integer;
  VTS,S,X : String;
begin
  x := substr(v,#13,'<CR>');
  x := substr(x,#10,'<LF>');
  VTS := inttostr(vt);
  B1 := chr(length(VTS)+32);
  B2 := chr(length(K)+32);
  B3 := chr((length(x) mod 200)+32);
  //R := 1+trunc(random(14));
  R := 0;                         //R=0 --> don't encrypt
  B4 := chr(R+32);
  S := B1+B2+B3+B4+VTS;
  For i := 1 to length(k) do begin
    C := ord(k[i]) xor R;
    S := S + chr(C);
  end;
  For i := 1 to length(x) do begin
    C := ord(x[i]) xor R;
    S := S + chr(C);
  end;
  result := S;
end;

Procedure UnPackRec(const S:string;var vt:integer;var k,v:string);
var
  L1,L2,L3,R,i : integer;
  C : Integer;
begin
  L1 := ord(S[1])-32;
  L2 := ord(S[2])-32;
  L3 := ord(S[3])-32;
  R  := ord(S[4])-32;

  vt := strtoint(copy(S,5,L1));   // type

  K := '';
  V := '';

  For i := 1 to L2 do begin
    C := ord(S[4+l1+i]) xor R;
    K := K + chr(C);
  end;

  // special case where we have more than 200 bytes of text
  if (4+L1+L2+L3) <> length(S) then L3 := length(S) - (4+L1+L2);

  For i := 1 to L3 do begin
    C := ord(S[4+L1+L2+i]) xor R;
    V := V + chr(C);
  end;
  v := substr(v,'<CR>',#13);
  v := substr(v,'<LF>',#10);
end;



Function StringFromVar(const v:Variant):string;
{ encode the variant as a string
  Handle dates properly}
var
  wv : variant;
  vt : integer;
begin
  wv := v;
  vt := vartype(wv);
  if vt = vardate then wv := varastype(wv,vardouble);
  result := varastype(wv,varstring);
end;


Function VarFromString(const vtype:Integer;const s:string):variant;
{ decode a string version of a variant
  Handle dates properly}
begin
  result := s;
  if vtype = vardate then begin
    Result := varastype(Result,varinteger);
    Result := varastype(result,vardate);
  end;
end;



Function TKeyDataFile.OpenForRead(const aFileName:String):integer;
{ open the file for reading
  return the number of data records }
begin
  Open := true;
  result := 0;
  if not fileexists(aFilename) then exit;
  assignfile(F,AFilename);
  reset(F);
  readln(F,nrecs);
  result := nrecs;
  if nrecs > 0 then reading := true;
end;

Procedure TKeyDataFile.OpenForWrite(const aFileName:String;const NoToWrite:integer);
begin
  Open := true;
  assignfile(F,AFilename);
  rewrite(F);
  writeln(F,NoToWrite);
  reading := false;
end;

Destructor TKeyDataFile.Destroy;
begin
  Close;
  inherited;
end;

Procedure TKeyDataFile.Close;
begin
  try closefile(F); except; end;    // silently deal with close file errors
  open := false;
end;

Procedure TKeyDataFile.WriteNext(const ky:string;v:variant);
begin
  writeln(F,PackRec(vartype(v),ky,StringFromVar(v)));
end;


Function TKeyDataFile.ReadNext(var ky:string;var v:variant):boolean;
var
  S,vs : string;
  vt : integer;
begin
  result := not eof(F);
  if result then begin
    readln(F,S);
    unpackrec(S,vt,ky,vs);
    v := VarFromString(vt,vs);
  end;
end;


end.
